/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Create
Version:   1.0
Status:    Stable
Date:      February 2002
ID:        ID_CREATE
Category:  Command
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2001-2002.  All rights reserved.
Keywords:  create, new, class, object

###DESCRIPTION###
<p>The Create class provides a means to dynamically create new objects during
script processing.  It is also used in graphical interfaces that require 
objects to be created when certain user events occur.</p>

<p>To create a new object with this class, you need to know the name of the
<a href="#Class">Class</a> that will be used in the creation process.  You may
set a <a href="#Target">Target</a> container for newly created objects, along
with optional <a href="#Flags">Flags</a> that affect the creation process.
Any fields that you want to set in the new object can be defined through the
unlisted fields mechanism.</p>

<p>The following example illustrates how to create a new Picture object:</p>

<pre>
    &lt;create class="picture" name="mypicture" &location="athene:pictures/screenshot.png"/&gt;
</pre>

<p>If necessary, a Create object can be defined as static for situations
where you need to create a new object during run-time operations.  You
can also use the Activate option to automatically activate a new object
when it is created.  For example:</p>

<pre>
   &lt;create static activate class="script" &src="$[text.input]$"/&gt;
</pre>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_CREATE 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_CREATE, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2001.  All rights reserved.",
   "July 2001",
   "Create"
};

static struct KernelBase  *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static OBJECTPTR StringsMod = 0;
static OBJECTPTR CreateClass = 0;
static FIELD FID_ObjectName;

/*****************************************************************************
** Class definition.
*/

#define MAX_FIELDS 30

struct Create {
   OBJECT_HEADER
   LONG Static;
   LONG Flags;
   OBJECTID TargetID;
   OBJECTID ObjectID;
   LONG Activate;
   LONG Execute;

   PRIVATE_FIELDS
   CLASSID ClassID;
   BYTE ObjectName[MAX_NAMELEN];
   MEMORYID Fields[MAX_FIELDS+1];
};

static ERROR GET_Class(struct Create *, struct Variable *);
static ERROR GET_ObjectName(struct Create *, STRING *);

static ERROR SET_Class(struct Create *, struct Variable *);
static ERROR SET_ObjectName(struct Create *, STRING);

static struct FieldDef CreateFlags[] = {
   { "NOTRACK", NF_NOTRACK }, { "PUBLIC",    NF_PUBLIC },
   { "SHARED",  NF_SHARED },  { "UNTRACKED", NF_UNTRACKED },
   { NULL, NULL }
};

static struct FieldArray CreateFields[] = {
   { "Static",   0, FDF_LONG|FDF_INIT,    0, NULL, NULL },
   { "Flags",    0, FDF_LONGFLAGS|FDF_RW, (LONG)&CreateFlags, NULL, NULL },
   { "Target",   0, FDF_OBJECTID|FDF_RW,  0, NULL, NULL },
   { "Object",   0, FDF_OBJECTID|FDF_R,   0, NULL, NULL },
   { "Activate", 0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Execute",  0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   /*** Virtual fields ***/
   { "Class",      0, FDF_VARIABLE|FDF_STRING|FDF_RW, 0, GET_Class, SET_Class },
   { "ObjectName", 0, FDF_STRING|FDF_RW,              0, GET_ObjectName, SET_ObjectName },
   END_FIELD
};

static ERROR CREATE_Activate(struct Create *, APTR);
static ERROR CREATE_ClosingTag(struct Create *, APTR);
static ERROR CREATE_Free(struct Create *, APTR);
static ERROR CREATE_Init(struct Create *, APTR);
static ERROR CREATE_SetUnlistedField(struct Create *, struct acSetUnlistedField *);

static struct ActionArray CreateActions[] = {
   { AC_Activate,         CREATE_Activate },
   { AC_ClosingTag,       CREATE_ClosingTag },
   { AC_Free,             CREATE_Free },
   { AC_Init,             CREATE_Init },
   { AC_SetUnlistedField, CREATE_SetUnlistedField },
   { NULL, NULL }
};

/*****************************************************************************
** User Instructions.
*/

static BYTE CreateUsage[] = {
"USAGE: CREATE CLASS=name [OBJECTNAME]=name [TARGET]=container [STATIC] ARGS..."
};

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL, FID_Name|TSTRING, "strings", TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   /*** Resolve field names ***/

   if (ResolveFields("ObjectName", &FID_ObjectName, TAGEND) != ERR_Okay) {
      return(ObjectError(argModule, ERH_InitModule, ERR_ResolveFields));
   }

   /*** Create the class ***/

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&CreateClass, NULL,
      FID_BaseClassID|TLONG, ID_CREATE,
      FID_Version|TFLOAT,    VER_CREATE,
      FID_Name|TSTRING,      "Create",
      FID_Category|TLONG,    CCF_COMMAND,
      FID_Usage|TPTR,        CreateUsage,
      FID_Actions|TPTR,      CreateActions,
      FID_Fields|TPTR,       CreateFields,
      FID_Size|TLONG,        sizeof(struct Create),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (StringsMod)  { Action(AC_Free, StringsMod, NULL);  StringsMod = NULL; }
   if (CreateClass) { Action(AC_Free, CreateClass, NULL); CreateClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Activates a Create object.
###END###

*****************************************************************************/

static ERROR CREATE_Activate(struct Create *Self, APTR Void)
{
   OBJECTPTR object, container;
   LONG i, j;
   BYTE str[1024], tmp[1024];
   STRING fieldname, fieldvalue;

   DPrintF("~Activate()","[Create:%d]", Self->Head.UniqueID);

   if (!Self->TargetID) {
      DPrintF("@Activate()","[Create:%d] The Target field has not been set.", Self->Head.UniqueID);
      StepBack();
      return(ERR_FieldNotSet);
   }

   if (NewObject(Self->ClassID, Self->Flags, &object, &Self->ObjectID) IS ERR_Okay) {

      /*** Set the object's container ***/

      if (AccessObject(Self->TargetID, 3000, &container) IS ERR_Okay) {
         SetOwner(object, container);
         ReleaseObject(container);
      }

      /*** Set the object's name ***/

      if (Self->ObjectName[0]) SetName(object, Self->ObjectName);

      /*** Set the object's fields ***/

      for (i=0; Self->Fields[i]; i++) {
         if (AccessMemory(Self->Fields[i], MEM_READ, (void **)&fieldname) IS ERR_Okay) {
            for (fieldvalue=fieldname; *fieldvalue; fieldvalue++);
            fieldvalue++;
            for (j=0; (fieldvalue[j]) AND (fieldvalue[j] != '['); j++);

            if (fieldvalue[j] IS '[') {
               StrCopy(fieldvalue, str, sizeof(str));
               StrTranslate(str, sizeof(str), NULL);
               if ((*str IS '=') AND (StrCalculate(str, NULL, tmp, sizeof(tmp)) IS ERR_Okay)) {
                  StrCopy(tmp, str, sizeof(str));
               }
               SetFieldVariable(object, fieldname, str);
            }
            else if ((*fieldvalue IS '=') AND (StrCalculate(fieldvalue, NULL, tmp, sizeof(tmp)) IS ERR_Okay)) {
               StrCopy(tmp, str, sizeof(str));
               SetFieldVariable(object, fieldname, str);
            }
            else SetFieldVariable(object, fieldname, fieldvalue);

            ReleaseMemory(fieldname);
         }
         else ObjectError(Self, ERH_Activate, ERR_AccessMemory);
      }

      if (Action(AC_Init, object, NULL) IS ERR_Okay) {

         /*** Automatic activation / Kill process ***/

         if (Self->Activate IS TRUE) {
            Action(AC_Activate, object, NULL);
            if (Self->Execute IS TRUE) Action(AC_Free, object, NULL);
         }
         else if (Self->Execute IS TRUE) {
            Action(AC_Activate, object, NULL);
            Action(AC_Free, object, NULL);
         }

         ReleaseObject(object);
         StepBack();
         return(ERR_Okay);
      }
      else {
         ReleaseObject(object);
         StepBack();
         return(ObjectError(Self, ERH_Activate, ERR_Init));
      }
   }
   else {
      StepBack();
      return(ObjectError(Self, ERH_Activate, ERR_NewObject));
   }
}

/*****************************************************************************

###ACTION###
Name:  ClosingTag
Short: If the object is non-static, this action will cause the Create object to activate and then self-destruct.
###END###

*****************************************************************************/

static ERROR CREATE_ClosingTag(struct Create *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      Action(AC_Activate, Self, NULL);
      Action(AC_Free, Self, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Create: Free
*/

static ERROR CREATE_Free(struct Create *Self, APTR Void)
{
   WORD i;

   for (i=0; Self->Fields[i]; i++) {
      FreeMemoryID(Self->Fields[i]);
      Self->Fields[i] = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Create: Init
*/

static ERROR CREATE_Init(struct Create *Self, APTR Void)
{
   if (!Self->ClassID) {
      DPrintF("@Init:","[Create:%d] The Class field has not been set.", Self->Head.UniqueID);
      return(ERR_FieldNotSet);
   }

   if (!Self->TargetID) {
      Self->TargetID = GetOwner(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTIONS###
Name:  SetUnlistedField
Short: This action is supported for setting new object field values.
###END###

*****************************************************************************/

static ERROR CREATE_SetUnlistedField(struct Create *Self, struct acSetUnlistedField *Args)
{
   WORD i, len;
   STRING fieldname;

   if ((!Args) OR (!Args->Field) OR (!Args->Value)) return(ERR_Args);

   for (i=0; Self->Fields[i]; i++) {
      if (AccessMemory(Self->Fields[i], MEM_READ, (void **)&fieldname) IS ERR_Okay) {
         if (StrCompare(Args->Field, fieldname, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            ReleaseMemory(fieldname);
            FreeMemoryID(Self->Fields[i]);
            Self->Fields[i] = NULL;
            break;
         }
         ReleaseMemory(fieldname);
      }
   }

   if (i < MAX_FIELDS) {
      if (AllocMemory(StrLength(Args->Field) + StrLength(Args->Value) + 2,
          MEM_STRING|MEM_NOCLEAR, (void **)&fieldname, &Self->Fields[i]) IS ERR_Okay) {

         if (Args->Field[0] IS '?') i = StrCopy(Args->Field+1, fieldname, COPY_ALL) + 1;
         else i = StrCopy(Args->Field, fieldname, COPY_ALL) + 1;

         StrCopy(Args->Value, fieldname + i, COPY_ALL);

         if (fieldname[i] IS '<') {
            len = StrLength(fieldname+i);
            if (fieldname[i+len-1] IS '>') {
               fieldname[i] = '[';
               fieldname[i+len-1] = ']';
            }
         }
         ReleaseMemory(fieldname);
      }
   }
   else return(ERR_Failed);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Activate
Short:  Set to TRUE if created objects should be automatically activated.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If you set the Activate field to TRUE, newly created objects will be
activated using the Activate action.  This behaviour is useful if you are
creating a new object for execution purposes.</p>

###SEE ALSO###
Field: Execute

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Class
Short:  Set this field to the name or ID of the class that the new object will be created from.
Type:   STRING/LONG
Status: Get/Set

###DESCRIPTION###
<p>You will need to set this field so that it refers to the class that will
be used when creating a new object.  If the Class field is not set or is
invalid, the Create object will fail when activated.</p>

<p>This field can be set with an integer that reflects a class ID, or a string
that reflects the name of a registered class.</p>

###END###

*****************************************************************************/

static ERROR GET_Class(struct Create *Self, struct Variable *Value)
{
   if (Value->Type & FD_LARGE) {
      Self->ClassID = (CLASSID)Value->Large;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
}

static ERROR SET_Class(struct Create *Self, struct Variable *Value)
{
   if (Value->Type & FD_LARGE) {
      Self->ClassID = (CLASSID)Value->Large;
      return(ERR_Okay);
   }
   else if (Value->Type & FD_STRING) {
      if ((Self->ClassID = ResolveClassName(Value->Pointer))) {
         return(ERR_Okay);
      }
      else {
         DPrintF("@Set.Class:","[Create] Failed to identify an ID for class \"%s\".", Value->Pointer);
         return(ERR_Failed);
      }
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
}

/*****************************************************************************

###FIELD###
Name:   Execute
Short:  Set to TRUE if created objects should be automatically killed.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If you set the Execute field to TRUE, newly created objects will be
activated using the Activate action, then they will be immediately freed
from the system.  This behaviour is useful if you do not require an object's
services after creating it.</p>

###SEE ALSO###
Field: Activate

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  This field contains optional flags that affect object creation.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>You can specify a number of optional flags in this field that affect
object creation.  These flags are identical to those listed in the object
kernel's <a href="../modules/kernel.html">NewObject()</a> function.  The
currently supported flags are:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td>NF_NOTRACK</td><td>An object created with this flag will not be tracked back to the object that created it.  If combined with the NF_PUBLIC flag, the object can also remain in memory when the object's related Task is terminated.</td></tr>
<tr><td>NF_PUBLIC</td><td>By using the NF_PUBLIC flag, you can allocate public objects rather than the default private type.  A public object can be accessed and interacted with by any application directly, because the object structure and its data lies in public memory.</td></tr>
<tr><td>NF_SHARED</td><td>Use this flag if you want to provide foreign tasks with a means of finding and talking to the created object.  If you have already specified the NF_PUBLIC field, it is not necessary to use the NF_SHARED flag, as all public objects have shared attributes.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Object
Short:  Reflects the ID of the last object that was created.
Type:   OBJECTID
Status: Read

###DESCRIPTION###
<p>After you have created a new object through the Activate action, you can
read the Object field to find out the new object's ID.  If you intend to
create more than one new object, you may want to store the values set in
this field as they will be over-written each time a new object is created.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ObjectName
Short:  This field determines the name that is given to newly created objects.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>If you want an object to be given a name at the moment of creation, set
this field to the name that it should be given.  Please note that there is a
limitation on the amount of characters available for naming; we recommend
that you do not exceed 25 characters in the name string.  Extraneous
characters will be cut short.</p>

###END###

*****************************************************************************/

static ERROR GET_ObjectName(struct Create *Self, STRING *Value)
{
   *Value = Self->ObjectName;
   return(ERR_Okay);
}

static ERROR SET_ObjectName(struct Create *Self, STRING Value)
{
   LONG i;

   if (Value) {
      for (i=0; (Value[i]) AND (i < MAX_NAMELEN-1); i++) Self->ObjectName[i] = Value[i];
      Self->ObjectName[i] = 0;
   }
   else Self->ObjectName[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Static
Short:  Set to TRUE to make a Create object static.
Type:   BOOLEAN
Status: Read/Init

###DESCRIPTION###
<p>To make a Create object static, set this field to TRUE.  If the Static
field is not set, then the object will self-destruct when a closing tag is
received.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Target
Short:  Defines the target container that new objects will be intialised to.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>It is highly recommended that you specify a target container before
creating new objects.  The object that you refer to in the Target field
will determine who owns a newly created object.  This will affect resource
tracking, so when the target object dies, the created object will die with
it.</p>

<p>If you do not set the Target field, new objects will be contained by
the container of the Create object.  This may or may not cause problems for
you, which is why we recommend setting the Target yourself.</p>

###END###

*****************************************************************************/
